using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Net;
using Microsoft.Xna.Framework.Storage;

namespace SillyText
{
    public class SillyChar
    {
        #region Internal properties

        private string charString;
        private Vector2 charPosition;
        private Color charColor;
        private Color sideColor;
        private Color shadowColor;
        private  Vector2 charOrigin;
        private float rotation=0;
        private float scale=1.0f;
        private  SpriteFont charFont;
        private float sideDistance;
        private float widthValue;
        private float heightValue;

        #endregion

        #region Public properties

        public float ShadowDistance = 100.0f;
        public float LightAngle = 0.0f;

        public float Width
        {
            get
            {
                return widthValue;
            }
        }

        public float Height
        {
            get
            {
                return heightValue;
            }
        }

        #endregion


        public SillyChar( string inCharString, SpriteFont inCharFont, Color inCharColor, Color inSideColor, Color inShadowColor, Vector2 inCharPosition)
        {
            charString = inCharString;
            charFont = inCharFont;
            charColor = inCharColor;
            sideColor = inSideColor;
            shadowColor = inShadowColor;
            Vector2 nSize = charFont.MeasureString("n");
            sideDistance = nSize.X / 5.0f;
            Vector2 charSize = inCharFont.MeasureString(charString);
            widthValue = charSize.X + sideDistance;
            heightValue = charSize.Y + sideDistance;
            charOrigin = new Vector2(charSize.X / 2.0f, charSize.Y / 2.0f);
            charPosition = new Vector2(inCharPosition.X+charOrigin.X, inCharPosition.Y+charOrigin.Y);
        }

        public void Draw ( SpriteBatch spriteBatch )
        {
            Vector2 drawPosition = new Vector2(charPosition.X + wanderX, charPosition.Y + wanderY);

            Vector2 sidePosition = new Vector2(
                (float)(sideDistance * (Math.Cos(LightAngle) / 2)),
                (float)(sideDistance * (Math.Sin(LightAngle) / 2))) + drawPosition;
            spriteBatch.DrawString(charFont, charString, sidePosition, sideColor, rotation, charOrigin, 1, SpriteEffects.None, 0.5f);
            Vector2 shadowPosition = new Vector2(
                (float)(ShadowDistance * (Math.Cos(LightAngle) / 2)),
                (float)(ShadowDistance * (Math.Sin(LightAngle) / 2))) + drawPosition;
            spriteBatch.DrawString(charFont, charString, shadowPosition, shadowColor, rotation, charOrigin, 1 + ShadowDistance/100.0f, SpriteEffects.None,0.5f);
            spriteBatch.DrawString(charFont, charString, drawPosition, charColor, rotation, charOrigin, scale, SpriteEffects.None, 0);
        }

        public delegate void UpdateSillyChar();

        private UpdateSillyChar updateBehavior;

        #region Simple Rotation Update

        private float rotateSpeed;
        public const float MaxRotateSpeed = 10;
        private bool isRotating;

        public bool StartRotate(float speed)
        {
            if (Math.Abs(speed) > MaxRotateSpeed)
            {
                return false;
            }
            rotateSpeed = speed;
            if (!isRotating)
            {
                updateBehavior += updateRotate;
                isRotating = true;
            }
            return true;
        }

        public void StopRotate()
        {
            if (isRotating)
            {
                updateBehavior -= updateRotate;
                isRotating = false;
            }
        }

        public float RotateSpeed 
        {
            get 
            {
                return rotateSpeed;
            }
            set 
            {
                if ( Math.Abs(value) <= MaxRotateSpeed) rotateSpeed = value;
            }
        }

        private const float twoPi = (float) (Math.PI * 2);
        private const float minusTwoPi = (float) (Math.PI * -2);

        private void updateRotate()
        {
            rotation += rotateSpeed;
            if (rotation > twoPi) rotation -= twoPi;
            if (rotation < minusTwoPi) rotation += twoPi;
        }

        #endregion

        #region Straighten

        private void updateStraighten()
        {
            if (Math.Abs(rotation) < .0001f)
            {
                rotation = 0;
                updateBehavior -= new UpdateSillyChar(updateStraighten);
            }
            else
            {
                rotation -= rotateSpeed;
            }
        }

        public void Straighten(int noOfTicks)
        {
            StopRotate();
            StopTwizzle();
            rotateSpeed = rotation / (float)noOfTicks;
            updateBehavior += updateStraighten;
        }

        #endregion

        #region Twizzle

        static Random twizzleRandom = new System.Random(1);

        private float twizzleRotateSpeedRange = 0.04f;
        private float twizzleRotateStepRange = 0.003f;
        private float twizzleRotateLow;
        private float twizzleRotateHigh;
        private float twizzleRotateStep;
        private float twizzleRotateSpeed;
        private bool isTwizzling;

        public void StartTwizzle()
        {
            twizzleRotateLow = -(twizzleRotateSpeedRange * (float)twizzleRandom.NextDouble());
            twizzleRotateHigh = (twizzleRotateSpeedRange * (float)twizzleRandom.NextDouble());
            twizzleRotateStep = twizzleRotateStepRange - (2 * (twizzleRotateStepRange * (float)twizzleRandom.NextDouble()));
            twizzleRotateSpeed = ((float)twizzleRandom.NextDouble() * (twizzleRotateHigh - twizzleRotateLow)) + twizzleRotateLow;

            if (!isTwizzling)
            {
                updateBehavior += updateTwizzle;
                isTwizzling = true;
            }
        }

        public void StopTwizzle()
        {
            if (isTwizzling)
            {
                updateBehavior -= new UpdateSillyChar(updateTwizzle);
                isTwizzling = false;
            }
        }

        private void updateTwizzle()
        {
            rotation += twizzleRotateSpeed;
            if (rotation > twoPi) rotation -= twoPi;
            if (rotation < minusTwoPi) rotation += twoPi;
            twizzleRotateSpeed += twizzleRotateStep;
            if (twizzleRotateSpeed > twizzleRotateHigh)
            {
                twizzleRotateStep = Math.Abs(twizzleRotateStep) * -1;
            }
            if (twizzleRotateSpeed < twizzleRotateLow)
            {
                twizzleRotateStep = Math.Abs(twizzleRotateStep);
            }
        }

        #endregion

        #region Wander

        public enum WanderStates
        {
            wandering,
            returning,
            off
        }

        private WanderStates wanderStateValue = WanderStates.off;

        public WanderStates WanderState
        {
            get
            {
                return wanderStateValue;
            }
            set
            {
                switch (value)
                {
                    case WanderStates.off:
                        switch (wanderStateValue)
                        {
                            case WanderStates.off:
                                break;
                            case WanderStates.wandering:
                            case WanderStates.returning:
                                wanderX = 0;
                                wanderY = 0;
                                updateBehavior -= new UpdateSillyChar(updateWander);
                                wanderStateValue = WanderStates.off;
                                break;
                        }
                        break;

                    case WanderStates.wandering:
                        switch (wanderStateValue)
                        {
                            case WanderStates.off:
                                wanderXLow = -(wanderSpeedRange * (float)wanderRandom.NextDouble());
                                wanderXHigh = -wanderXLow; // (wanderSpeedRange * (float)wanderRandom.NextDouble());
                                wanderXStep = wanderStepRange - (2 * (wanderStepRange * (float)wanderRandom.NextDouble()));
                                wanderXSpeed = 0;// ((float)wanderRandom.NextDouble() * (wanderXHigh - wanderXLow)) + wanderXLow;

                                wanderYLow = -(wanderSpeedRange * (float)wanderRandom.NextDouble());
                                wanderYHigh = -wanderYLow; // (wanderSpeedRange * (float)wanderRandom.NextDouble());
                                wanderYStep = wanderStepRange - (2 * (wanderStepRange * (float)wanderRandom.NextDouble()));
                                wanderYSpeed = 0;// ((float)wanderRandom.NextDouble() * (wanderYHigh - wanderYLow)) + wanderYLow;

                                wanderX = 0;
                                wanderY = 0;

                                updateBehavior += updateWander;

                                wanderStateValue = WanderStates.wandering;
                                break;

                            case WanderStates.wandering:
                                break;

                            case WanderStates.returning:
                                wanderStateValue = WanderStates.wandering;
                                break;
                        }
                        
                        break;

                    case WanderStates.returning:
                        switch (wanderStateValue)
                        {
                            case WanderStates.off:
                                break;
                            case WanderStates.wandering:
                                wanderStateValue = WanderStates.returning;
                                break;
                            case WanderStates.returning:
                                break;
                        }
                        break;
                }
            }
        }

        static Random wanderRandom = new System.Random(1);
        private float wanderSpeedRange = 5.0f;
        private float wanderStepRange = 0.5f;

        private float wanderXLow;
        private float wanderXHigh;
        private float wanderXStep;
        private float wanderXSpeed;

        private float wanderYLow;
        private float wanderYHigh;
        private float wanderYStep;
        private float wanderYSpeed;
        
        float wanderX;
        float wanderY;

        private void updateWander()
        {
            switch (WanderState)
            {
                case WanderStates.off:
                    break;
                case WanderStates.wandering:
                    wanderX += wanderXSpeed;
                    wanderXSpeed += wanderXStep;
                    if (wanderXSpeed > wanderXHigh)
                    {
                        wanderXStep = Math.Abs(wanderXStep) * -1;
                    }
                    if (wanderXSpeed < wanderXLow)
                    {
                        wanderXStep = Math.Abs(wanderXStep);
                    }

                    wanderY += wanderYSpeed;
                    wanderYSpeed += wanderYStep;
                    if (wanderYSpeed > wanderYHigh)
                    {
                        wanderYStep = Math.Abs(wanderYStep) * -1;
                    }
                    if (wanderYSpeed < wanderYLow)
                    {
                        wanderYStep = Math.Abs(wanderYStep);
                    }
                    break;

                case WanderStates.returning:
                    wanderX += wanderXSpeed;
                    wanderXSpeed += wanderXStep;
                    if (wanderXSpeed > wanderXHigh)
                    {
                        wanderXStep = Math.Abs(wanderXStep) * -1;
                    }
                    if (wanderXSpeed < wanderXLow)
                    {
                        wanderXStep = Math.Abs(wanderXStep);
                    }

                    wanderY += wanderYSpeed;
                    wanderYSpeed += wanderYStep;
                    if (wanderYSpeed > wanderYHigh)
                    {
                        wanderYStep = Math.Abs(wanderYStep) * -1;
                    }
                    if (wanderYSpeed < wanderYLow)
                    {
                        wanderYStep = Math.Abs(wanderYStep);
                    }

                    if ( ((wanderX*wanderX) + (wanderY*wanderY)) < 4 ) {
                        wanderX = 0;
                        wanderY = 0;
                        WanderState = WanderStates.off;
                    }
                    break;
            }
        }

        #endregion

        #region Return

        private bool isReturning = false;

        private void updateReturn()
        {
            if ( (wanderX < 0 && wanderXSpeed > 0) || (wanderX > 0 && wanderXSpeed < 0) ||
                 (wanderY < 0 && wanderYSpeed > 0) || (wanderY > 0 && wanderYSpeed < 0))
            {
                wanderX = 0;
                wanderY = 0;
                updateBehavior -= new UpdateSillyChar(updateReturn);
                isReturning = false;
            }
            else
            {
                wanderX -= wanderXSpeed;
                wanderY -= wanderYSpeed;
            }
        }

        public void returnToPosition(int noOfTicks)
        {
            if (!isReturning)
            {
                wanderXSpeed = wanderX / (float)noOfTicks;
                wanderYSpeed = wanderY / (float)noOfTicks;
                updateBehavior += updateReturn;
                isReturning = true;
            }
        }

        #endregion

        public void Update()
        {
            if (updateBehavior != null)
            {
                updateBehavior();
            }
        }
    }

    public class SillyLine
    {
        SillyChar[] line;

        public void Update()
        {
            foreach (SillyChar ch in line)
            {
                ch.Update();
            }
        }

        public void Draw(SpriteBatch spriteBatch)
        {
            foreach (SillyChar ch in line)
            {
                ch.Draw(spriteBatch);
            }
        }

        public Vector2 Position;

        private Random rand = new Random(1);

        public SillyLine(
            string text,
            Vector2 inPosition,
            Color[] charColors,
            Color[] sideColors, 
            Color[] shadowColors,
            SpriteFont font )
        {
            Position = inPosition;

            Vector2 drawPosition = inPosition;
            int sideColorPos = 0;
            int shadowColorPos = 0;
            int colorPos = 0;

            line = new SillyChar[text.Length];

            for (int i = 0; i < text.Length; i++)
            {

                SillyChar word = new SillyChar (
                    text[i].ToString(),
                    font,
                    charColors[colorPos],
                    sideColors[sideColorPos],
                    shadowColors[shadowColorPos],
                    drawPosition);

                line[i] = word;

                drawPosition.X += line[i].Width;

                if (++colorPos == charColors.Length) colorPos = 0;
                if (++sideColorPos == sideColors.Length) sideColorPos = 0;
                if (++shadowColorPos == shadowColors.Length) shadowColorPos = 0;
            }
        }

        public void StartTwizzle()
        {
            foreach (SillyChar ch in line)
            {
                ch.StartTwizzle();
            }
        }

        public void StopTwizzle()
        {
            foreach (SillyChar ch in line)
            {
                ch.StopTwizzle();
            }
        }

        public SillyChar.WanderStates WanderState
        {
            get
            {
                foreach (SillyChar ch in line)
                {
                    if (ch.WanderState != SillyChar.WanderStates.off)
                    {
                        return SillyChar.WanderStates.wandering;
                    }
                }
                return SillyChar.WanderStates.off;
            }
            set
            {
                foreach (SillyChar ch in line)
                {
                    ch.WanderState = value;
                }
            }
        }

        //public void StartWander()
        //{
        //    foreach (SillyChar ch in line)
        //    {
        //        ch.StartWander();
        //    }
        //}

        //public void StopWander()
        //{
        //    foreach (SillyChar ch in line)
        //    {
        //        ch.StopWander();
        //    }
        //}

        public void Straighten(int ticks)
        {
            foreach (SillyChar ch in line)
            {
                ch.Straighten(ticks);
            }
        }

        public void Straighten(int minTicks, int maxTicks)
        {
            foreach (SillyChar ch in line)
            {
                ch.Straighten(rand.Next(minTicks, maxTicks));
            }
        }

        public void returnToPosition(int ticks)
        {
            foreach (SillyChar ch in line)
            {
                ch.returnToPosition(ticks);
            }
        }

        public void returnToPosition(int minTicks, int maxTicks)
        {
            foreach (SillyChar ch in line)
            {
                ch.returnToPosition(rand.Next(minTicks, maxTicks));
            }
        }

        public float LightAngle
        {
            set
            {
                foreach (SillyChar ch in line)
                {
                    ch.LightAngle = value;
                }
            }
        }

        public float ShadowDistance
        {
            set
            {
                foreach (SillyChar ch in line)
                {
                    ch.ShadowDistance = value;
                }
            }
        }
    }
}
