using System;
using System.Collections.Generic;
using System.Text;

using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Net;
using Microsoft.Xna.Framework.Storage;

namespace SillyText
{
    public class SillyWord : SillyText.ISillyText
    {
        #region Internal properties

        private string charString;
        private Vector2 charPosition;
        private Color charColor;
        private Color sideColor;
        private Color shadowColor;
        private Vector2 charOrigin;
        private SpriteFont charFont;
        private float sideDistance;
        private float widthValue;
        private float heightValue;

        private float x;
        private float y;

        private float xModify;
        private float yModify;
        private float rotateModify;
        private float sizeModify = 1.0f;
        private float rawScale = 1.0f;

        private const float twoPi = (float)(Math.PI * 2);
        private const float minusTwoPi = (float)(Math.PI * -2);

        #endregion

        #region Public properties

        public float X
        {
            get
            {
                return x;
            }
        }

        public float Y
        {
            get
            {
                return y;
            }
        }

        private float shadowDistanceValue = 10.0f;

        public float ShadowDistance
        {
            set
            {
                shadowDistanceValue = value;
            }
        }

        private float lightAngleValue = (float)Math.PI / 4;

        public float LightAngle
        {
            set
            {
                lightAngleValue = value;
            }
        }

        public float Width
        {
            get
            {
                return widthValue;
            }
        }

        public float Height
        {
            get
            {
                return heightValue;
            }
        }

        public Color TextColor
        {
            set
            {
                charColor = value;
            }
        }

        public Color SideColor
        {
            set
            {
                sideColor = value;
            }
        }

        public Color ShadowColor
        {
            set
            {
                shadowColor = value;
            }
        }

        #endregion

        public SillyWord(
            string inCharString, 
            SpriteFont inCharFont, 
            Color inCharColor, 
            Color inSideColor, 
            Color inShadowColor, 
            Vector2 inCharPosition,
            float nWidth)
        {
            x = inCharPosition.X;
            y = inCharPosition.Y;

            charString = inCharString;
            charFont = inCharFont;
            charColor = inCharColor;
            sideColor = inSideColor;
            shadowColor = inShadowColor;
            Vector2 nSize = charFont.MeasureString("n");
            rawScale = nWidth / nSize.X;
            sideDistance = nSize.X / 5.0f;
            Vector2 charSize = inCharFont.MeasureString(charString);
            widthValue = charSize.X * rawScale;
            heightValue = charSize.Y * rawScale;
            charOrigin = new Vector2(charSize.X / 2.0f, charSize.Y / 2.0f);
            charPosition = new Vector2(inCharPosition.X + charOrigin.X, inCharPosition.Y + charOrigin.Y);
        }

        public void Draw(SpriteBatch spriteBatch)
        {

            Vector2 drawPosition = new Vector2(charPosition.X + xModify, charPosition.Y + yModify);

            if (shadowDistanceValue > 0)
            {
                Vector2 sidePosition = new Vector2(
                    (float)(sideDistance * (Math.Cos(lightAngleValue) / 2)),
                    (float)(sideDistance * (Math.Sin(lightAngleValue) / 2))) + drawPosition;
                spriteBatch.DrawString(charFont, charString, sidePosition, sideColor, rotateModify, charOrigin, sizeModify, SpriteEffects.None, 0.5f);
                Vector2 shadowPosition = new Vector2(
                    (float)(shadowDistanceValue * (Math.Cos(lightAngleValue) / 2) * rawScale),
                    (float)(shadowDistanceValue * (Math.Sin(lightAngleValue) / 2))* rawScale) + drawPosition;
                spriteBatch.DrawString(charFont, charString, shadowPosition, shadowColor, rotateModify, charOrigin, (1 + shadowDistanceValue / 100.0f) * sizeModify, SpriteEffects.None, 0.5f);
            }
            spriteBatch.DrawString(charFont, charString, drawPosition, charColor, rotateModify, charOrigin, sizeModify, SpriteEffects.None, 0);
        }

        public bool IsMoving()
        {
            if (TwizzleState != TwizzleStates.off)
            {
                return true;
            }

            if (WanderState != WanderStates.off)
            {
                return true;
            }

            if (FlyInState == FlyInStates.flyingIn || FlyInState == FlyInStates.flyingOut)
            {
                return true;
            }
            return false;
        }

        #region Twizzle

        public enum TwizzleStates
        {
            twizzling,
            returning,
            off
        }

        private TwizzleStates twizzleStateValue = TwizzleStates.off;

        private void twizzleOff()
        {
            switch (twizzleStateValue)
            {
                case TwizzleStates.off:
                    break;
                case TwizzleStates.twizzling:
                case TwizzleStates.returning:
                    twizzleRotation = 0;
                    twizzleStateValue = TwizzleStates.off;
                    break;
            }
        }

        public TwizzleStates TwizzleState
        {
            get
            {
                return twizzleStateValue;
            }
            set
            {
                switch (value)
                {
                    case TwizzleStates.off:
                        twizzleOff();
                        break;

                    case TwizzleStates.twizzling:
                        switch (twizzleStateValue)
                        {
                            case TwizzleStates.twizzling:
                                break;
                            case TwizzleStates.returning:
                                twizzleStateValue = TwizzleStates.twizzling;
                                break;
                            case TwizzleStates.off:
                                twizzleRotateLow = -(twizzleRotateSpeedRange * (float)twizzleRandom.NextDouble() + twizzleRotateSpeedLow);
                                twizzleRotateHigh = (twizzleRotateSpeedRange * (float)twizzleRandom.NextDouble() + twizzleRotateSpeedLow);
                                twizzleRotateStep = twizzleRotateStepRange - (2 * (twizzleRotateStepRange * (float)twizzleRandom.NextDouble()));
                                twizzleRotateSpeed = ((float)twizzleRandom.NextDouble() * (twizzleRotateHigh - twizzleRotateLow)) + twizzleRotateLow;
                                twizzleStateValue = TwizzleStates.twizzling;
                                break;
                        }
                        break;

                    case TwizzleStates.returning:
                        switch (twizzleStateValue)
                        {
                            case TwizzleStates.twizzling:
                                twizzleRotateSpeed = twizzleRotateHigh;
                                twizzleStateValue = TwizzleStates.returning;
                                break;
                            case TwizzleStates.returning:
                                break;
                            case TwizzleStates.off:
                                break;
                        }
                        break;
                }
            }
        }

        public void StartTwizzle(
            float inTwizzleRotateLow,
            float inTwizzleRotateHigh,
            float inTwizzleRotateStep,
            float inTwizzleRorateSpeed)
        {
            twizzleRotateLow = inTwizzleRotateLow;
            twizzleRotateHigh = inTwizzleRotateHigh;
            twizzleRotateStep = inTwizzleRotateStep;
            twizzleRotateSpeed = inTwizzleRorateSpeed;
            twizzleStateValue = TwizzleStates.twizzling;
        }

        static Random twizzleRandom = new System.Random(1);

        private float twizzleRotateSpeedLow = 0.01f;
        private float twizzleRotateSpeedRange = 0.04f;
        private float twizzleRotateStepRange = 0.003f;
        private float twizzleRotateLow;
        private float twizzleRotateHigh;
        private float twizzleRotateStep;
        private float twizzleRotateSpeed;
        private float twizzleRotation = 0;

        private void twizzleRotate()
        {
            twizzleRotation += twizzleRotateSpeed;
            if (twizzleRotation > twoPi) twizzleRotation -= twoPi;
            if (twizzleRotation < minusTwoPi) twizzleRotation += twoPi;
            rotateModify += twizzleRotation;
        }

        private void updateTwizzle()
        {
            switch (twizzleStateValue)
            {
                case TwizzleStates.off:
                    break;
                case TwizzleStates.twizzling:
                    twizzleRotate();
                    twizzleRotateSpeed += twizzleRotateStep;
                    if (twizzleRotateSpeed > twizzleRotateHigh)
                    {
                        twizzleRotateStep = Math.Abs(twizzleRotateStep) * -1;
                    }
                    if (twizzleRotateSpeed < twizzleRotateLow)
                    {
                        twizzleRotateStep = Math.Abs(twizzleRotateStep);
                    }
                    break;
                case TwizzleStates.returning:
                    twizzleRotate();
                    if (Math.Abs(twizzleRotation) < Math.Abs(2 * twizzleRotateSpeed))
                    {
                        TwizzleState = TwizzleStates.off;
                    }
                    break;
            }
        }

        #endregion

        #region Wander

        static Random wanderRandom = new System.Random(1);
        private float wanderSpeedLow = 2.0f;
        private float wanderSpeedRange = 2.0f;
        private float wanderStepRange = 0.5f;

        private float wanderXLow;
        private float wanderXHigh;
        private float wanderXStep;
        private float wanderXSpeed;

        private float wanderYLow;
        private float wanderYHigh;
        private float wanderYStep;
        private float wanderYSpeed;

        float wanderX;
        float wanderY;

        public enum WanderStates
        {
            wandering,
            returning,
            off
        }

        private WanderStates wanderStateValue = WanderStates.off;

        public WanderStates WanderState
        {
            get
            {
                return wanderStateValue;
            }
            set
            {
                switch (value)
                {
                    case WanderStates.off:
                        switch (wanderStateValue)
                        {
                            case WanderStates.off:
                                break;
                            case WanderStates.wandering:
                            case WanderStates.returning:
                                wanderX = 0;
                                wanderY = 0;
                                wanderStateValue = WanderStates.off;
                                break;
                        }
                        break;

                    case WanderStates.wandering:
                        switch (wanderStateValue)
                        {
                            case WanderStates.off:
                                wanderXLow = -(wanderSpeedRange * (float)wanderRandom.NextDouble()) - wanderSpeedLow;
                                wanderXHigh = -wanderXLow; // (wanderSpeedRange * (float)wanderRandom.NextDouble());
                                wanderXStep = wanderStepRange - (2 * (wanderStepRange * (float)wanderRandom.NextDouble()));
                                wanderXSpeed = 0;// ((float)wanderRandom.NextDouble() * (wanderXHigh - wanderXLow)) + wanderXLow;

                                wanderYLow = -(wanderSpeedRange * (float)wanderRandom.NextDouble()) - wanderSpeedLow;
                                wanderYHigh = -wanderYLow; // (wanderSpeedRange * (float)wanderRandom.NextDouble());
                                wanderYStep = wanderStepRange - (2 * (wanderStepRange * (float)wanderRandom.NextDouble()));
                                wanderYSpeed = 0;// ((float)wanderRandom.NextDouble() * (wanderYHigh - wanderYLow)) + wanderYLow;

                                wanderX = 0;
                                wanderY = 0;

                                wanderStateValue = WanderStates.wandering;
                                break;

                            case WanderStates.wandering:
                                break;

                            case WanderStates.returning:
                                wanderStateValue = WanderStates.wandering;
                                break;
                        }

                        break;

                    case WanderStates.returning:
                        switch (wanderStateValue)
                        {
                            case WanderStates.off:
                                break;
                            case WanderStates.wandering:
                                wanderStateValue = WanderStates.returning;
                                break;
                            case WanderStates.returning:
                                break;
                        }
                        break;
                }
            }
        }

        private void updateWander()
        {
            switch (WanderState)
            {
                case WanderStates.off:
                    break;
                case WanderStates.wandering:
                    wanderX += wanderXSpeed;
                    wanderXSpeed += wanderXStep;
                    if (wanderXSpeed > wanderXHigh)
                    {
                        wanderXStep = Math.Abs(wanderXStep) * -1;
                    }
                    if (wanderXSpeed < wanderXLow)
                    {
                        wanderXStep = Math.Abs(wanderXStep);
                    }

                    wanderY += wanderYSpeed;

                    wanderYSpeed += wanderYStep;
                    if (wanderYSpeed > wanderYHigh)
                    {
                        wanderYStep = Math.Abs(wanderYStep) * -1;
                    }
                    if (wanderYSpeed < wanderYLow)
                    {
                        wanderYStep = Math.Abs(wanderYStep);
                    }

                    break;

                case WanderStates.returning:
                    wanderX += wanderXSpeed;
                    wanderXSpeed += wanderXStep;
                    if (wanderXSpeed > wanderXHigh)
                    {
                        wanderXStep = Math.Abs(wanderXStep) * -1;
                    }
                    if (wanderXSpeed < wanderXLow)
                    {
                        wanderXStep = Math.Abs(wanderXStep);
                    }

                    wanderY += wanderYSpeed;
                    wanderYSpeed += wanderYStep;
                    if (wanderYSpeed > wanderYHigh)
                    {
                        wanderYStep = Math.Abs(wanderYStep) * -1;
                    }
                    if (wanderYSpeed < wanderYLow)
                    {
                        wanderYStep = Math.Abs(wanderYStep);
                    }

                    if (((wanderX * wanderX) + (wanderY * wanderY)) < 4)
                    {
                        wanderX = 0;
                        wanderY = 0;
                        WanderState = WanderStates.off;
                    }

                    break;
            }
            xModify += wanderX;
            yModify += wanderY;
        }

        #endregion

        #region Fly In

        static Random flyInRandom = new System.Random(1);

        private float flyInXSpeed;
        private float flyInYSpeed;

        private int minFlyInTime = 60;
        private int maxFlyInTime = 120;
        private int minDistance = 1000;
        private int maxDistance = 1200;

        private float flyScale = 1;
        private float flyScaleStep;
        private float flyScaleLow = 0.05f;
        private float flyScaleHigh = 10.0f;

        float flyInX;
        float flyInY;

        public enum FlyInStates
        {
            flyingIn,
            flyingOut,
            flownOut,
            flownIn,
            moving
        }

        private FlyInStates flyInStateValue = FlyInStates.flownIn;

        private void setRandomPosition()
        {
            flyInX = flyInRandom.Next(minDistance, maxDistance);
            if (flyInRandom.Next(2) == 0)
            {
                flyInX = -flyInX;
            }

            flyInY = flyInRandom.Next(minDistance, maxDistance);
            if (flyInRandom.Next(2) == 0)
            {
                flyInY = -flyInY;
            }

            flyScale = (float) (flyScaleLow + (flyInRandom.NextDouble() * (flyScaleHigh - flyScaleLow)));
        }

        private void calculateFlySpeeds(int ticks)
        {
            flyInXSpeed = (float)(flyInX / ticks);
            flyInYSpeed = (float)(flyInY / ticks);
            flyScaleStep = (1.0f - flyScale) / ticks;
        }

        public void StartFlyIn(int minTicks, int maxTicks)
        {
            StartFlyIn(flyInRandom.Next(minTicks, maxTicks));
        }

        public void StartFlyIn(int ticks)
        {
            setRandomPosition();
            calculateFlySpeeds(ticks);
            flyInStateValue = FlyInStates.flyingIn;
        }

        private void startFlyingIn()
        {
            switch (flyInStateValue)
            {
                case FlyInStates.flownIn:
                case FlyInStates.flownOut:
                    setRandomPosition();
                    calculateFlySpeeds(flyInRandom.Next(minFlyInTime, maxFlyInTime));
                    flyInStateValue = FlyInStates.flyingIn;
                    break;

                case FlyInStates.flyingIn:
                    break;

                case FlyInStates.flyingOut:
                    calculateFlySpeeds(flyInRandom.Next(minFlyInTime, maxFlyInTime));
                    flyInStateValue = FlyInStates.flyingIn;
                    break;
            }
        }

        public void StartFlyOut(int minTicks, int maxTicks)
        {
            StartFlyOut(flyInRandom.Next(minTicks, maxTicks));
        }

        public void StartFlyOut(int ticks)
        {
            setRandomPosition();
            calculateFlySpeeds(ticks);
            flyInX = 0;
            flyInY = 0;
            flyScale = 1;
            flyInStateValue = FlyInStates.flyingOut;
        }

        private void startFlyingOut()
        {
            switch (flyInStateValue)
            {
                case FlyInStates.flownIn:
                case FlyInStates.flownOut:
                    StartFlyOut(flyInRandom.Next(minFlyInTime, maxFlyInTime));
                    break;

                case FlyInStates.flyingIn:
                    calculateFlySpeeds(flyInRandom.Next(minFlyInTime, maxFlyInTime));
                    flyInStateValue = FlyInStates.flyingOut;
                    break;

                case FlyInStates.flyingOut:
                    break;
            }
        }

        public FlyInStates FlyInState
        {
            get
            {
                return flyInStateValue;
            }
            set
            {
                switch (value)
                {
                    case FlyInStates.moving:
                        break;

                    case FlyInStates.flownOut:
                        setRandomPosition();
                        flyInStateValue = FlyInStates.flownOut;
                        break;

                    case FlyInStates.flownIn:
                        flyInX = 0;
                        flyInY = 0;
                        flyInStateValue = FlyInStates.flownIn;
                        break;

                    case FlyInStates.flyingIn:
                        startFlyingIn();
                        break;

                    case FlyInStates.flyingOut:
                        startFlyingOut();
                        break;
                }
            }
        }

        private void updateFlyIn()
        {
            switch (FlyInState)
            {
                case FlyInStates.flownOut:
                    break;
                case FlyInStates.flownIn:
                case FlyInStates.moving:
                    break;

                case FlyInStates.flyingIn:
                    flyInX -= flyInXSpeed;
                    flyInY -= flyInYSpeed;
                    flyScale += flyScaleStep;
                    if (((flyInX * flyInX) + (flyInY * flyInY)) < 4)
                    {
                        flyInX = 0;
                        flyInY = 0;
                        FlyInState = FlyInStates.flownIn;
                    }
                    break;

                case FlyInStates.flyingOut:
                    flyInX += flyInXSpeed;
                    flyInY += flyInYSpeed;
                    flyScale -= flyScaleStep;
                    if (Math.Abs(flyInX) > maxDistance && Math.Abs(flyInY) > maxDistance)
                    {
                        FlyInState = FlyInStates.flownOut;
                    }

                    break;
            }
            xModify += flyInX;
            yModify += flyInY;
            sizeModify *= flyScale;
        }

        #endregion

        #region Offset

        private float offsetXValue = 0;
        private float offsetYValue = 0;

        public float OffsetX
        {
            set
            {
                offsetXValue = value;
            }
            get
            {
                return offsetXValue;
            }
        }

        public float OffsetY
        {
            set
            {
                offsetYValue = value;
            }
            get
            {
                return offsetYValue;
            }
        }

        #endregion

        public void Update()
        {
            rotateModify = 0;
            xModify = offsetXValue;
            yModify = offsetYValue;
            sizeModify = rawScale;

            updateWander();
            updateTwizzle();
            updateFlyIn();
        }
    }

    public class SillyLine : SillyText.ISillyText
    {
        protected SillyWord[] line;

        public float X
        {
            get
            {
                return Position.X;
            }
        }

        public float Y
        {
            get
            {
                return Position.Y;
            }
        }

        float widthValue;
        float heightValue;

        public float Width
        {
            get
            {
                return widthValue;
            }
        }

        public float Height
        {
            get
            {
                return heightValue;
            }
        }

        public void Update()
        {
            foreach (SillyWord ch in line)
            {
                ch.Update();
            }
        }

        public void Draw(SpriteBatch spriteBatch)
        {
            foreach (SillyWord ch in line)
            {
                ch.Draw(spriteBatch);
            }
        }

        public bool IsMoving()
        {
            foreach (SillyWord ch in line)
            {
                if (ch.IsMoving())
                {
                    return true;
                }
            }
            return false;
        }

        private Vector2 Position;

        private Random rand = new Random(1);

        public SillyLine(
            string text,
            Vector2 inPosition,
            Color[] charColors,
            Color[] sideColors,
            Color[] shadowColors,
            SpriteFont font,
            float nWidth
            )
        {
            Position = inPosition;

            Vector2 drawPosition = inPosition;
            heightValue = 0;
            int sideColorPos = 0;
            int shadowColorPos = 0;
            int colorPos = 0;

            line = new SillyWord[text.Length];

            for (int i = 0; i < text.Length; i++)
            {

                SillyWord word = new SillyWord(
                    text[i].ToString(),
                    font,
                    charColors[colorPos],
                    sideColors[sideColorPos],
                    shadowColors[shadowColorPos],
                    drawPosition,
                    nWidth);

                line[i] = word;

                drawPosition.X += line[i].Width;
                if ( line[i].Height > heightValue ) 
                {
                    heightValue = line[i].Height;
                }

                if (++colorPos == charColors.Length) colorPos = 0;
                if (++sideColorPos == sideColors.Length) sideColorPos = 0;
                if (++shadowColorPos == shadowColors.Length) shadowColorPos = 0;
            }
            widthValue = drawPosition.X - inPosition.X;
        }

        #region Twizzle

        public SillyWord.TwizzleStates TwizzleState
        {
            get
            {
                foreach (SillyWord ch in line)
                {
                    if (ch.TwizzleState != SillyWord.TwizzleStates.off)
                    {
                        return SillyWord.TwizzleStates.twizzling;
                    }
                }
                return SillyWord.TwizzleStates.off;
            }
            set
            {
                foreach (SillyWord ch in line)
                {
                    ch.TwizzleState = value;
                }
            }

        }

        public void StartTwizzle(
            float inTwizzleRotateLow,
            float inTwizzleRotateHigh,
            float inTwizzleRotateStep,
            float inTwizzleRorateSpeed)
        {
            foreach (SillyWord ch in line)
            {
                ch.StartTwizzle(inTwizzleRotateLow, inTwizzleRotateHigh, inTwizzleRotateStep, inTwizzleRorateSpeed);
            }
        }

        #endregion

        #region Wander

        public SillyWord.WanderStates WanderState
        {
            get
            {
                foreach (SillyWord ch in line)
                {
                    if (ch.WanderState != SillyWord.WanderStates.off)
                    {
                        return SillyWord.WanderStates.wandering;
                    }
                }
                return SillyWord.WanderStates.off;
            }
            set
            {
                foreach (SillyWord ch in line)
                {
                    ch.WanderState = value;
                }
            }
        }


        #endregion

        #region Fly In and Fly Out

        public SillyWord.FlyInStates FlyInState
        {
            get
            {
                if (line.Length > 0)
                {
                    SillyWord.FlyInStates result = line[0].FlyInState;
                    foreach (SillyWord ch in line)
                    {
                        if (ch.FlyInState != result)
                        {
                            return SillyWord.FlyInStates.moving;
                        }
                    }
                    return result;
                }
                return SillyWord.FlyInStates.flownIn;
            }

            set
            {
                foreach (SillyWord ch in line)
                {
                    ch.FlyInState = value;
                }
            }
        }

        public void StartFlyIn(int minTicks, int maxTicks)
        {
            foreach (SillyWord ch in line)
            {
                ch.StartFlyIn(minTicks, maxTicks);
            }
        }

        public void StartFlyIn(int ticks)
        {
            foreach (SillyWord ch in line)
            {
                ch.StartFlyIn(ticks);
            }
        }

        public void StartFlyOut(int minTicks, int maxTicks)
        {
            foreach (SillyWord ch in line)
            {
                ch.StartFlyOut(minTicks, maxTicks);
            }
        }

        public void StartFlyOut(int ticks)
        {
            foreach (SillyWord ch in line)
            {
                ch.StartFlyOut(ticks);
            }
        }

        #endregion

        #region Rendering

        public float OffsetX
        {
            set
            {
                foreach (SillyWord ch in line)
                {
                    ch.OffsetX = value;
                }
            }
            get
            {
                if (line.Length > 0)
                {
                    return line[0].OffsetX;
                }
                else
                {
                    return 0;
                }
            }
        }

        public float OffsetY
        {
            set
            {
                foreach (SillyWord ch in line)
                {
                    ch.OffsetY = value;
                }
            }
            get
            {
                if (line.Length > 0)
                {
                    return line[0].OffsetY;
                }
                else
                {
                    return 0;
                }
            }
        }

        public float LightAngle
        {
            set
            {
                foreach (SillyWord ch in line)
                {
                    ch.LightAngle = value;
                }
            }
        }

        public float ShadowDistance
        {
            set
            {
                foreach (SillyWord ch in line)
                {
                    ch.ShadowDistance = value;
                }
            }
        }

        public Color TextColor
        {
            set
            {
                foreach (SillyWord ch in line)
                {
                    ch.TextColor = value;
                }
            }
        }

        public Color SideColor
        {
            set
            {
                foreach (SillyWord ch in line)
                {
                    ch.SideColor = value;
                }
            }
        }

        public Color ShadowColor
        {
            set
            {
                foreach (SillyWord ch in line)
                {
                    ch.ShadowColor = value;
                }
            }
        }

        #endregion

    }

    public class SillyHtmlLine : SillyText.ISillyText
    {
        List<SillyWord> line = new List<SillyWord>();

        private Vector2 startPosition;

        public float X
        {
            get
            {
                return startPosition.X;
            }
        }

        public float Y
        {
            get
            {
                return startPosition.Y;
            }
        }

        float widthValue;
        float heightValue;

        public float Width
        {
            get
            {
                return widthValue;
            }
        }

        public float Height
        {
            get
            {
                return heightValue;
            }
        }

        string htmlText;

        int textPos;

        Color drawColor;

        public void Update()
        {
            foreach (SillyWord ch in line)
            {
                ch.Update();
            }
        }

        public void Draw(SpriteBatch spriteBatch)
        {
            foreach (SillyWord ch in line)
            {
                ch.Draw(spriteBatch);
            }
        }

        public bool IsMoving()
        {
            foreach (SillyWord ch in line)
            {
                if (ch.IsMoving())
                {
                    return true;
                }
            }
            return false;
        }

        private string getNextText()
        {
            StringBuilder result = new StringBuilder();

            bool inDirective = false;
            bool inControl = false;

            while (true)
            {
                if (textPos == htmlText.Length)
                {
                    if (result.Length == 0)
                    {
                        return "";
                    }
                    else
                    {
                        return result.ToString();
                    }
                }

                char nextCh = htmlText[textPos];

                if (char.IsControl(nextCh))
                {
                    textPos++;
                    continue;
                }

                if (nextCh == '&')
                {
                    if (result.Length > 0)
                    {
                        return result.ToString();
                    }
                    if (inControl)
                    {
                        throw new System.Exception("Nested & characters");
                    }
                    inControl = true;
                }

                if (nextCh == '<')
                {
                    if (result.Length > 0)
                    {
                        return result.ToString();
                    }

                    // spaces in a directive do not end it
                    if (inDirective)
                    {
                        throw new System.Exception("Nested < characters");
                    }
                    inDirective = true;
                }

                textPos++;

                if (nextCh == ' ' & !inDirective)
                {
                    if (result.Length > 0)
                    {
                        result.Append(' ');
                        return result.ToString();
                    }
                }
                else
                {
                    result.Append(nextCh);
                }

                if (nextCh == '>')
                {
                    return result.ToString();
                }

                if (nextCh == ';' && inControl)
                {
                    return result.ToString();
                }
            }
        }

        public SillyHtmlLine(
            string inHtmlText,
            Vector2 inPosition,
            Color charColor,
            Color sideColor,
            Color highlightColor,
            Color shadowColor,
            SpriteFont font,
            float nWidth,
            int maxX)
        {
            startPosition = inPosition;
            drawColor = charColor;

            Vector2 drawPosition = startPosition;

            htmlText = inHtmlText;
            textPos = 0;

            Vector2 nSize = font.MeasureString("n");
            float rawScale = nWidth / nSize.X;

            float singleLineSpace = font.MeasureString("M").Y * rawScale;
            float spaceWidth = font.MeasureString(" ").X * rawScale;
            float hyphenWidth = font.MeasureString("-").X * rawScale;
            float drawWidth = maxX - inPosition.X;

            while (true)
            {
                if (drawPosition.X > widthValue) widthValue = drawPosition.X;
                string nextText = getNextText();

                if (nextText.Length == 0)
                {
                    break;
                }

                if (nextText[0] == '<')
                {
                    // got an HTML directive
                    nextText = nextText.ToUpper();

                    switch (nextText)
                    {
                        case @"<P>":
                            break;
                        case @"</P>":
                            drawPosition.X = startPosition.X;
                            drawPosition.Y += (singleLineSpace * 1.5f);
                            break;
                    }
                }
                else
                {
                    // just plain text
                    string nextDirective = nextText.ToLower();

                    if (nextDirective.StartsWith("&"))
                    {
                        switch (nextDirective)
                        {
                            case @"&nbsp;":
                                nextText = " ";
                                break;
                            case @"&quot;":
                                nextText = "\"";
                                break;
                            case @"&reg;":
                                nextText = "(r)";
                                break;
                        }
                        if (nextText.Length > 0)
                        {
                            if (drawPosition.X + font.MeasureString(nextText).X > maxX)
                            {
                                drawPosition.X = startPosition.X;
                                drawPosition.Y += singleLineSpace;
                            }
                            foreach (char ch in nextText)
                            {
                                SillyWord s = new SillyWord(
                                    ch.ToString(),
                                    font,
                                    drawColor,
                                    sideColor,
                                    shadowColor,
                                    drawPosition,
                                    nWidth);
                                drawPosition.X += s.Width;
                                line.Add(s);
                            }
                        }
                    }
                    else
                    {
                        if (nextText.Length > 0)
                        {
                            // need to draw something
                            if (font.MeasureString(nextText).X > drawWidth)
                            {
                                // if the string is larger then the display width 
                                // display each letter in turn and break with a hyphen each time

                                // start at a new line
                                drawPosition.X = startPosition.X;
                                drawPosition.Y += singleLineSpace;

                                foreach (char ch in nextText)
                                {
                                    SillyWord s = new SillyWord(
                                        ch.ToString(),
                                        font,
                                        drawColor,
                                        sideColor,
                                        shadowColor,
                                        drawPosition,
                                        nWidth);

                                    if ((drawPosition.X + s.Width + hyphenWidth) > maxX)
                                    {
                                        // need to break the word
                                        SillyWord hyphen = new SillyWord(
                                            "-",
                                            font,
                                            drawColor,
                                            sideColor,
                                            shadowColor,
                                            drawPosition,
                                            nWidth);
                                        line.Add(hyphen);
                                        drawPosition.X = startPosition.X;
                                        drawPosition.Y += singleLineSpace;
                                        // redraw the letter at the start of the line
                                        s = new SillyWord(
                                            ch.ToString(),
                                            font,
                                            drawColor,
                                            sideColor,
                                            shadowColor,
                                            drawPosition,
                                            nWidth);
                                    }
                                    // move down the line
                                    drawPosition.X += s.Width;
                                    // add the letter
                                    line.Add(s);
                                }
                            }
                            else
                            {
                                if (drawPosition.X + font.MeasureString(nextText).X > maxX)
                                {
                                    drawPosition.X = startPosition.X;
                                    drawPosition.Y += singleLineSpace;
                                }
                                foreach (char ch in nextText)
                                {
                                    SillyWord s = new SillyWord(
                                        ch.ToString(),
                                        font,
                                        drawColor,
                                        sideColor,
                                        shadowColor,
                                        drawPosition,
                                        nWidth);
                                    drawPosition.X += s.Width;
                                    line.Add(s);
                                }
                            }
                        }
                    }
                }
            }

            widthValue -= inPosition.X;
            heightValue = (drawPosition.Y + singleLineSpace) - inPosition.Y;
        }

        #region Twizzle

        public SillyWord.TwizzleStates TwizzleState
        {
            get
            {
                foreach (SillyWord ch in line)
                {
                    if (ch.TwizzleState != SillyWord.TwizzleStates.off)
                    {
                        return SillyWord.TwizzleStates.twizzling;
                    }
                }
                return SillyWord.TwizzleStates.off;
            }
            set
            {
                foreach (SillyWord ch in line)
                {
                    ch.TwizzleState = value;
                }
            }

        }

        public void StartTwizzle(
            float inTwizzleRotateLow,
            float inTwizzleRotateHigh,
            float inTwizzleRotateStep,
            float inTwizzleRorateSpeed)
        {
            foreach (SillyWord ch in line)
            {
                ch.StartTwizzle(inTwizzleRotateLow, inTwizzleRotateHigh, inTwizzleRotateStep, inTwizzleRorateSpeed);
            }
        }

        #endregion

        #region Wander

        public SillyWord.WanderStates WanderState
        {
            get
            {
                foreach (SillyWord ch in line)
                {
                    if (ch.WanderState != SillyWord.WanderStates.off)
                    {
                        return SillyWord.WanderStates.wandering;
                    }
                }
                return SillyWord.WanderStates.off;
            }
            set
            {
                foreach (SillyWord ch in line)
                {
                    ch.WanderState = value;
                }
            }
        }


        #endregion

        #region Fly In and Fly Out

        public SillyWord.FlyInStates FlyInState
        {
            get
            {
                if (line.Count > 0)
                {
                    SillyWord.FlyInStates result = line[0].FlyInState;
                    foreach (SillyWord ch in line)
                    {
                        if (ch.FlyInState != result)
                        {
                            return SillyWord.FlyInStates.moving;
                        }
                    }
                    return result;
                }
                return SillyWord.FlyInStates.flownIn;
            }

            set
            {
                foreach (SillyWord ch in line)
                {
                    ch.FlyInState = value;
                }
            }
        }

        public void StartFlyIn(int minTicks, int maxTicks)
        {
            foreach (SillyWord ch in line)
            {
                ch.StartFlyIn(minTicks, maxTicks);
            }
        }

        public void StartFlyIn(int ticks)
        {
            foreach (SillyWord ch in line)
            {
                ch.StartFlyIn(ticks);
            }
        }

        public void StartFlyOut(int minTicks, int maxTicks)
        {
            foreach (SillyWord ch in line)
            {
                ch.StartFlyOut(minTicks, maxTicks);
            }
        }

        public void StartFlyOut(int ticks)
        {
            foreach (SillyWord ch in line)
            {
                ch.StartFlyOut(ticks);
            }
        }

        #endregion

        #region Rendering

        public float OffsetX
        {
            set
            {
                foreach (SillyWord ch in line)
                {
                    ch.OffsetX = value;
                }
            }
            get
            {
                if (line.Count > 0)
                {
                    return line[0].OffsetX;
                }
                else
                {
                    return 0;
                }
            }

        }

        public float OffsetY
        {
            set
            {
                foreach (SillyWord ch in line)
                {
                    ch.OffsetY = value;
                }
            }
            get
            {
                if (line.Count > 0)
                {
                    return line[0].OffsetY;
                }
                else
                {
                    return 0;
                }
            }

        }

        public float LightAngle
        {
            set
            {
                foreach (SillyWord ch in line)
                {
                    ch.LightAngle = value;
                }
            }
        }

        public float ShadowDistance
        {
            set
            {
                foreach (SillyWord ch in line)
                {
                    ch.ShadowDistance = value;
                }
            }
        }

        public Color TextColor
        {
            set
            {
                foreach (SillyWord ch in line)
                {
                    ch.TextColor = value;
                }
            }
        }

        public Color SideColor
        {
            set
            {
                foreach (SillyWord ch in line)
                {
                    ch.SideColor = value;
                }
            }
        }

        public Color ShadowColor
        {
            set
            {
                foreach (SillyWord ch in line)
                {
                    ch.ShadowColor = value;
                }
            }
        }


        #endregion

    }
}
